@@ -1,6 +1,6 @@
 /*
  *	cfg.c - configuration file stuff
- *	Copyright (C) 2000, 2001 Fred Barnes <frmb2@ukc.ac.uk>
+ *	Copyright (C) 2000-2003 Fred Barnes <frmb2@ukc.ac.uk>
  *
  *	This program is free software; you can redistribute it and/or modify
  *	it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 
 #define __CFG_C
 
+/*{{{  includes*/
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -38,14 +39,16 @@
 #include "cfg.h"
 #include "interface.h"
 #include "error.h"
-
-
-static FILE *config_stream;
+/*}}}*/
+/*{{{  statics, externs */
 #define NUM_CONFIG_FILES (2)
 static char *config_files[] = {
 	SYSCONFDIR "/xdmchoose.conf",
 	"./xdmchoose.conf"
 };
+extern char *cmdline_conffile;			/* in chooser.c */
+extern int grabkybd;				/* in chooser.c */
+
 static char *config_file = NULL;
 static int config_fd = -1;
 static char *config_mmap = NULL;
@@ -56,14 +59,29 @@ static int config_nlines = 0;
 static int data_line;
 /*          %SERVER%      %SERVER.IP% %XDMHOST%      %XDMHOST.IP% */
 static char *server_name, *server_ip, *xdmhost_name, *xdmhost_ip;
-
-
+/*}}}*/
+/*{{{  forward decls*/
 static int scanto (char *ident, char *setting);
 static int extract_int (char *ident, char *setting, int *dest);
 static int extract_mstring (char *ident, char *setting, char **dest);
+static int extract_lstrings (char *ident, char *setting, char ***dest);
+/*}}}*/
+/*{{{  void show_config_filenames (FILE *stream)*/
+/*
+ *	void show_config_filenames (FILE *stream)
+ *	shows the various compiled-in config filenames, one per line, tab indented.
+ */
+void show_config_filenames (FILE *stream)
+{
+	int i;
 
-
-
+	for (i=0; i<NUM_CONFIG_FILES; i++) {
+		fprintf (stream, "\t%s\n", config_files[i]);
+	}
+	return;
+}
+/*}}}*/
+/*{{{  int init_config_file (GlobalConfigData *data)*/
 /*
  *	int init_config_file (GlobalConfigData *data)
  *	Initialises the configuration stuff, placing global settings in 'data'
@@ -119,16 +137,24 @@ int init_config_file (GlobalConfigData *data)
 			xdmhost_ip = string_dup (inet_ntoa (*(struct in_addr *)(hp->h_addr_list[0])));
 		}
 	}
-	for (i=0; i<NUM_CONFIG_FILES; i++) {
-		if (!access (config_files[i], R_OK)) {
-			break;
+	if (cmdline_conffile) {
+		if (access (cmdline_conffile, R_OK)) {
+			APP_MSG ("unable to read config file %s", cmdline_conffile);
+			return -1;
 		}
+		config_file = cmdline_conffile;
+	} else {
+		for (i=0; i<NUM_CONFIG_FILES; i++) {
+			if (!access (config_files[i], R_OK)) {
+				break;
+			}
+		}
+		if (i == NUM_CONFIG_FILES) {
+			APP_MSG ("unable to find valid configuration file");
+			return -1;
+		}
+		config_file = config_files[i];
 	}
-	if (i == NUM_CONFIG_FILES) {
-		APP_MSG ("unable to find valid configuration file");
-		return -1;
-	}
-	config_file = config_files[i];
 	if (stat (config_file, &st_buf)) {
 		APP_MSG ("unable to stat configuration file %s", config_file);
 		return -1;
@@ -207,6 +233,9 @@ int init_config_file (GlobalConfigData *data)
 	data->ctrlcexitcode = 255;
 	data->stripdomain = 0;
 	data->startscriptkill = NULL;
+	data->allowtype = 1;
+	data->hostlist = NULL;
+	data->grabkybd = grabkybd;
 	extract_int ("app", "hidequery", &data->hidequery);
 	extract_int ("app", "allowbroadcast", &data->allowbroadcast);
 	extract_mstring ("app", "startscript", &data->startscript);
@@ -215,6 +244,9 @@ int init_config_file (GlobalConfigData *data)
 	extract_int ("app", "ctrlcexitcode", &data->ctrlcexitcode);
 	extract_int ("app", "stripdomain", &data->stripdomain);
 	extract_mstring ("app", "startscriptkill", &data->startscriptkill);
+	extract_int ("app", "allowtype", &data->allowtype);
+	extract_lstrings ("app", "hostlist", &data->hostlist);
+	extract_int ("app", "grabkeyboard", &data->grabkybd);
 	if (data->stripdomain) {
 		/* remove .XYZ.foo.bar from server_name and xdmhost_name */
 		char *ch;
@@ -230,8 +262,8 @@ int init_config_file (GlobalConfigData *data)
 	}
 	return 0;
 }
-
-
+/*}}}*/
+/*{{{  void close_config_file (void)*/
 /*
  *	void close_config_file (void)
  *	closes the config file
@@ -253,8 +285,8 @@ void close_config_file (void)
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  int read_config_entry (char *item, GadgetConfigData *data, int fields)*/
 /*
  *	int read_config_entry (char *item, GadgetConfigData *data, int fields)
  *	fills up 'data' with config information about 'item'
@@ -323,8 +355,41 @@ int read_config_entry (char *item, GadgetConfigData *data, int fields)
 	}
 	return errored;
 }
+/*}}}*/
+/*{{{  int have_config_entry (char *item, int fields)*/
+/*
+ *	checks to see if particular entries are here
+ *	return 0 on success, number of non-matches on failure
+ */
+int have_config_entry (char *item, int fields)
+{
+	int errored = 0;
 
-
+	if ((fields & FLD_TEXT) && scanto (item, "text")) {
+		errored++;
+	}
+	if ((fields & FLD_FOREGROUND) && scanto (item, "foreground")) {
+		errored++;
+	}
+	if ((fields & FLD_BACKGROUND) && scanto (item, "background")) {
+		errored++;
+	}
+	if ((fields & FLD_HIGHLIGHT) && scanto (item, "highlight")) {
+		errored++;
+	}
+	if ((fields & FLD_SHADOW) && scanto (item, "shadow")) {
+		errored++;
+	}
+	if ((fields & FLD_FONT) && scanto (item, "font")) {
+		errored++;
+	}
+	if ((fields & FLD_TENTRY) && scanto (item, "tentry")) {
+		errored++;
+	}
+	return errored;
+}
+/*}}}*/
+/*{{{  static int scanto (char *ident, char *setting)*/
 /*
  *	int scanto (char *ident, char *setting)
  *	scans to a particular line in the config file
@@ -363,8 +428,8 @@ static int scanto (char *ident, char *setting)
 	}
 	return 0;
 }
-
-
+/*}}}*/
+/*{{{  static int extract_int (char *ident, char *setting, int *dest)*/
 /*
  *	int extract_int (char *ident, char *setting, int *dest)
  *	Extracts an integer value of the specified name
@@ -386,8 +451,8 @@ static int extract_int (char *ident, char *setting, int *dest)
 	*dest = tmp;
 	return 0;
 }
-
-
+/*}}}*/
+/*{{{  static int extract_mstring (char *ident, char *setting, char **dest)*/
 /*
  *	int extract_mstring (char *ident, char *setting, char **dest)
  *	Extracts a string, but allocates memory for the return
@@ -396,9 +461,10 @@ static int extract_int (char *ident, char *setting, int *dest)
 static int extract_mstring (char *ident, char *setting, char **dest)
 {
 	char *ch;
-	char *sublist[4] = { "%SERVER%", "%SERVER.IP%", "%XDMHOST%", "%XDMHOST.IP%" };
-	char **withlist[4] = { &server_name, &server_ip, &xdmhost_name, &xdmhost_ip };
-	int num_items = 4;
+	char *sublist[5] = { "%SERVER%", "%SERVER.IP%", "%XDMHOST%", "%XDMHOST.IP%", "%VERSION%" };
+	static char *verstr = VERSION;
+	char **withlist[5] = { &server_name, &server_ip, &xdmhost_name, &xdmhost_ip, &verstr };
+	int num_items = 5;
 	int i, x, y, z, newlen, offs;
 	static char databuf[256];
 
@@ -446,5 +512,63 @@ static int extract_mstring (char *ident, char *setting, char **dest)
 	}
 	return 0;
 }
+/*}}}*/
+/*{{{  static int extract_lstrings (char *ident, char *setting, char ***dest)*/
+/*
+ *	static int extract_lstrings (char *ident, char *setting, char ***dest)
+ *	Extracts a list of strings, no subsitutions here
+ */
+static int extract_lstrings (char *ident, char *setting, char ***dest)
+{
+	static char databuf[1024];
+	char **tmp;
+	int n_tmp, m_tmp;
 
+	if (scanto (ident, setting)) {
+		if (scanto ("unknown", setting)) {
+			return -1;
+		}
+	}
+	if (strlen (data_start) > 1023) {
+		return -1;
+	}
+	strcpy (databuf, data_start);
+	data_start = databuf;
+	n_tmp = m_tmp = 0;
+	tmp = NULL;
+	while (*data_start == '"') {
+		char *ch;
+
+		for (ch = data_start + 1; (*ch != '\"') && (*ch != '\0'); ch++);
+		if (*ch == '\0') {
+			APP_MSG ("missing quote in config file for %s.%s in %s(%d).", ident, setting, config_file, data_line);
+			return -1;
+		}
+		*ch = '\0';
+		if (n_tmp == m_tmp) {
+			tmp = (char **)xrealloc (tmp, m_tmp * sizeof(char *), (m_tmp + 10) * sizeof(char *));
+			m_tmp += 10;
+		}
+		tmp[n_tmp++] = string_dup (data_start + 1);
+		for (ch++; (*ch == ' ') || (*ch == '\t'); ch++);
+		if (*ch == ',') {
+			for (ch++; (*ch == ' ') || (*ch == '\t'); ch++);
+		}
+		data_start = ch;
+	}
+	if (*data_start != '\0') {
+		APP_MSG ("garbage in config file for %s.%s in %s(%d).", ident, setting, config_file, data_line);
+		return -1;
+	}
+	data_start++;
+	/* sort out target */
+	if (n_tmp) {
+		*dest = (char **)xmalloc ((n_tmp + 1) * sizeof(char *));
+		memcpy (*dest, tmp, n_tmp * sizeof(char *));
+		(*dest)[n_tmp] = NULL;
+		free (tmp);
+	}
+	return 0;
+}
+/*}}}*/
 
