@@ -55,7 +55,7 @@ in this Software without prior written authorization from the X Consortium.
  */
 
 
-
+/*{{{  includes, defines*/
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -118,14 +118,8 @@ in this Software without prior written authorization from the X Consortium.
 #endif /* hpux */
 
 #include    <netdb.h>
-
-
-static ARRAY8Ptr xdmAddress;
-static ARRAY8Ptr  clientAddress;
-static int connectionType;
-static int child_pid;
-
-
+/*}}}*/
+/*{{{  structs*/
 typedef struct _hostAddr {
 	struct sockaddr *addr;
 	int addrlen;
@@ -139,6 +133,12 @@ typedef struct _hostName {
 	CARD16 connectionType;
 	ARRAY8 hostaddr;
 } HostName;
+/*}}}*/
+/*{{{  static vars*/
+static ARRAY8Ptr xdmAddress;
+static ARRAY8Ptr  clientAddress;
+static int connectionType;
+static int child_pid;
 
 static HostAddr sel_hostaddr;
 static HostName sel_hostname;
@@ -147,8 +147,9 @@ static int socketFD;
 static int pingTry;
 static int prog_state;
 
-enum PROG_STATES { NOTHING, QUERYING, CONNECTING };
+/*}}}*/
 
+enum PROG_STATES { NOTHING, QUERYING, CONNECTING };
 
 #define PING_INTERVAL	2000
 #define TRIES		3
@@ -177,6 +178,13 @@ extern int XdmcpAllocARRAY8 (ARRAY8Ptr, int);
 
 extern GlobalConfigData _globconfig;
 
+
+/* global */
+char *progname = NULL;
+char *cmdline_conffile = NULL;
+int grabkybd = 1;
+
+
 /*
  *	int query_host (void)
  *	Queries a host to get info about it
@@ -593,7 +601,7 @@ static void Choose (void)
 				exit (REMANAGE_DISPLAY);
 			}
 			sleep(5);	/* Hack because sometimes the connection gets
-					   closed before the data arrives on the other end. */
+					   closed before the data arrives on the other end.  well, i think.. */
 			t_snddis (fd,NULL);
 			t_unbind (fd);
 			t_close (fd);
@@ -622,6 +630,8 @@ static void host_select_function (char *name)
 	char buffer[256], tbuf[128], *ch, *dh;
 	struct hostent *hp;
 
+	sprintf (buffer, "resolving %s...", name);
+	set_status_message (buffer);
 	if (!name) {
 		set_error_message ("invalid hostname");
 		return;
@@ -674,6 +684,8 @@ static void host_check_function (char *name)
 	if (_globconfig.hidequery) {
 		return;
 	}
+	sprintf (buffer, "resolving %s...", name);
+	set_status_message (buffer);
 	if (!name) {
 		set_error_message ("invalid hostname");
 		return;
@@ -721,12 +733,35 @@ static void host_check_function (char *name)
 static void dump_version_info (FILE *stream)
 {
 	fprintf (stream, "xdmchoose (chooser) version " XDMCHOOSE_VERSION "\n");
-	fprintf (stream, "Copyright (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>\n");
+	fprintf (stream, "Copyright (C) 2000-2003 Fred Barnes <frmb2@ukc.ac.uk>\n");
 	fprintf (stream, "Portions copyright (C) MIT Consortium\n");
 	return;
 }
 
 
+/*
+ *	void dump_usage_info (FILE *stream)
+ *	dumps usage information
+ */
+static void dump_usage_info (FILE *stream)
+{
+	fprintf (stream, "xdmchoose (chooser) version " XDMCHOOSE_VERSION "\n");
+	fprintf (stream, "Usage: %s [options]\n", progname);
+	fprintf (stream, "where options are:\n");
+	fprintf (stream, "\t-xdmaddress <addr>          set XDM host address\n");
+	fprintf (stream, "\t-clientaddress <addr>       set client address\n");
+	fprintf (stream, "\t-connectiontype <c>         set the connection type (integer)\n");
+	fprintf (stream, "\t--nograbkybd | -g           don\'t grab the keyboard\n");
+	fprintf (stream, "\t--config | -c  <file>       specify config file to use\n");
+	fprintf (stream, "\t--version | -V              print version and exit\n");
+	fprintf (stream, "\t--help | -h                 this help\n");
+	fprintf (stream, "If no configuration file is specified on the command line (normal\n");
+	fprintf (stream, "when launched from xdm), the compiled-in defaults are:\n");
+	show_config_filenames (stream);
+	return;
+}
+
+
 /*
  *	void cleanup_child (void)
  *	cleans-up the child (script) process
@@ -763,15 +798,16 @@ int main (int argc, char **argv)
 	clientAddress = (ARRAY8Ptr)NULL;
 	connectionType = 0;
 	child_pid = 0;
+	for (progname = *argv + (strlen (*argv) - 1); (progname > *argv) && (progname[-1] != '/'); progname--);
 	for (walk = argv+1, i = 1; *walk && (i < argc); i++, walk++) {
 		if (!strcasecmp (*walk, "-xdmaddress")) {
 			walk++, i++;
 			if (!*walk || (i==argc)) {
-				fprintf (stderr, "%s: -xdmaddress missing argument\n", *argv);
+				fprintf (stderr, "%s: -xdmaddress missing argument\n", progname);
 				exit (EXIT_FAILURE);
 			}
 			if (StringToARRAY8 (&xdmAddress, *walk)) {
-				fprintf (stderr, "%s: unable to convert [%s]\n", *argv, *walk);
+				fprintf (stderr, "%s: unable to convert [%s]\n", progname, *walk);
 				exit (EXIT_FAILURE);
 			}
 			continue;
@@ -779,11 +815,11 @@ int main (int argc, char **argv)
 		if (!strcasecmp (*walk, "-clientaddress")) {
 			walk++, i++;
 			if (!*walk || (i == argc)) {
-				fprintf (stderr, "%s: -clientaddress missing argument\n", *argv);
+				fprintf (stderr, "%s: -clientaddress missing argument\n", progname);
 				exit (EXIT_FAILURE);
 			}
 			if (StringToARRAY8 (&clientAddress, *walk)) {
-				fprintf (stderr, "%s: unable to convert [%s]\n", *argv, *walk);
+				fprintf (stderr, "%s: unable to convert [%s]\n", progname, *walk);
 				exit (EXIT_FAILURE);
 			}
 			continue;
@@ -791,19 +827,36 @@ int main (int argc, char **argv)
 		if (!strcasecmp (*walk, "-connectiontype")) {
 			walk++, i++;
 			if (!*walk || (i == argc)) {
-				fprintf (stderr, "%s: -connectiontype missing argument\n", *argv);
+				fprintf (stderr, "%s: -connectiontype missing argument\n", progname);
 				exit (EXIT_FAILURE);
 			}
 			if (!sscanf (*walk, "%d", &connectionType)) {
-				fprintf (stderr, "%s: badly formed integer for -connectiontype [%s]\n", *argv, *walk);
+				fprintf (stderr, "%s: badly formed integer for -connectiontype [%s]\n", progname, *walk);
 				exit (EXIT_FAILURE);
 			}
 			continue;
 		}
-		if (!strcasecmp (*walk, "--version")) {
+		if (!strcmp (*walk, "--config") || !strcmp (*walk, "-c")) {
+			walk++, i++;
+			if (!*walk || (i == argc)) {
+				fprintf (stderr, "%s: %s missing argument\n", progname, walk[-1]);
+				exit (EXIT_FAILURE);
+			}
+			cmdline_conffile = *walk;
+			continue;
+		}
+		if (!strcmp (*walk, "--nograbkybd") || !strcmp (*walk, "-g")) {
+			grabkybd = 0;
+			continue;
+		}
+		if (!strcmp (*walk, "--version") || !strcmp (*walk, "-V")) {
 			dump_version_info (stderr);
 			exit (EXIT_SUCCESS);
 		}
+		if (!strcmp (*walk, "--help") || !strcmp (*walk, "-h")) {
+			dump_usage_info (stderr);
+			exit (EXIT_FAILURE);
+		}
 		/* Just ignore other command line things... */
 	}
 	if (init_xparts (argv, argc)) {
