@@ -1,6 +1,6 @@
 /*
  *	gadgets.c - gadgets implementation
- *	Copyright (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>
+ *	Copyright (C) 2000-2003 Fred Barnes <frmb2@ukc.ac.uk>
  *
  *	This program is free software; you can redistribute it and/or modify
  *	it under the terms of the GNU General Public License as published by
@@ -17,7 +17,9 @@
  *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+/* Much of this is based on code from WG:  http://frmb.org/wg.html */
 
+/*{{{  includes/defines*/
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -29,16 +31,20 @@
 #include <X11/Xutil.h>
 #include <X11/keysym.h>
 #include <X11/keysymdef.h>
-#include <X11/xpm.h>
+#include <X11/cursorfont.h>
 
 #include "support.h"
 #include "cfg.h"
 #include "interface.h"
 #include "error.h"
 
-#define MAX_LOOKUP_STRING (64)
+#ifdef USE_LIBXPM
+#include <X11/xpm.h>
+#endif
 
-/* structures and stuff */
+#define MAX_LOOKUP_STRING (64)
+/*}}}*/
+/*{{{  structures and stuff */
 typedef struct {
 	int text_x, text_y;
 	char *text;
@@ -53,6 +59,7 @@ typedef struct {
 	int cur_len, max_len, cur_pos;
 	unsigned long te_pixel;
 	int exposed;
+	int allow_type;
 	void (*edit_accept)(Gadget *, char *);
 	void (*edit_escape)(Gadget *, char *);
 	void (*edit_ctrlc)(Gadget *, char *);
@@ -61,13 +68,28 @@ typedef struct {
 typedef struct {
 	int text_x, text_y;
 	char *text;
+	int doarrow;
+	XPoint arrow_points[3];
 	int len;
 	int mouse_in, mouse_down;
 	int exposed;
 	void (*button_press)(Gadget *);
 } ButtonPrivate;
 
-/* prototypes */
+typedef struct {
+	char **options;
+	int n_opts;
+	int exposed;
+	int text_height;
+	int text_yoffs;
+	void (*list_select)(Gadget *, int);
+	int pop_x, pop_y;
+	int grabbed;
+	Cursor g_cur;			/* grabbed cursor */
+	Gadget *below;			/* thing to pop up under */
+} ListPrivate;
+/*}}}*/
+/*{{{  prototypes */
 static Gadget *create_gadget (MyXContext *cxt, char *name, int x, int y, int width, int height);
 static void destroy_gadget (Gadget *gadget);
 static void allocate_color (MyXContext *cxt, char *colordef, unsigned long *xpixel);
@@ -83,8 +105,12 @@ static void button_mousedown_event (XEvent *ev, Gadget *gadget);
 static void button_mouseup_event (XEvent *ev, Gadget *gadget);
 static void button_mouseenter_event (XEvent *ev, Gadget *gadget);
 static void button_mouseleave_event (XEvent *ev, Gadget *gadget);
-
-
+static void list_expose_event (XEvent *ev, Gadget *gadget);
+static void list_mousedown_event (XEvent *ev, Gadget *gadget);
+static void list_mouseup_event (XEvent *ev, Gadget *gadget);
+static void list_recalc_xypos (Gadget *gadget);
+/*}}}*/
+/*{{{  static Gadget *create_gadget (MyXContext *cxt, char *name, int x, int y, int width, int height)*/
 /*
  *	Gadget *create_gadget (MyXContext *cxt, char *name, int x, int y, int width, int height)
  *	Creates a new gadget
@@ -121,8 +147,8 @@ static Gadget *create_gadget (MyXContext *cxt, char *name, int x, int y, int wid
 	gadget->xcxt = cxt;
 	return gadget;
 }
-
-
+/*}}}*/
+/*{{{  static void destroy_gadget (Gadget *gadget)*/
 /*
  *	void destroy_gadget (Gadget *gadget)
  *	Destroys a gadget
@@ -138,8 +164,8 @@ static void destroy_gadget (Gadget *gadget)
 	free (gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  Gadget *create_toplevel (MyXContext *cxt, char *name, int x, int y, int width, int height)*/
 /*
  *	Gadget *create_toplevel (MyXContext *cxt, char *name, int x, int y, int width, int height)
  *	Creates a new top-level gadget
@@ -147,6 +173,12 @@ static void destroy_gadget (Gadget *gadget)
 Gadget *create_toplevel (MyXContext *cxt, char *name, int x, int y, int width, int height)
 {
 	Gadget *gadget;
+	XWMHints wm_hints;
+	XClassHint class_hints;
+	XSizeHints size_hints;
+	XTextProperty x_caption, x_icon_name;
+	XTextProperty *xc_ptr, *xi_ptr;
+	char *caption, *icon_name, *app_name, *app_class;
 
 	gadget = create_gadget (cxt, name, x, y, width, height);
 	if (read_config_entry (name, gadget->config, FLD_BACKGROUND)) {
@@ -155,11 +187,37 @@ Gadget *create_toplevel (MyXContext *cxt, char *name, int x, int y, int width, i
 	allocate_color (cxt, gadget->config->background, &gadget->bg_pixel);
 	gadget->window = XCreateSimpleWindow (cxt->display, RootWindow (cxt->display, cxt->screen),
 		x, y, width, height, 0, BlackPixel (cxt->display, cxt->screen), gadget->bg_pixel);
+	XSelectInput (cxt->display, gadget->window, KeyPressMask | KeyReleaseMask | StructureNotifyMask | FocusChangeMask | ExposureMask);
+	caption = string_dup ("XDM-Choose " VERSION);
+	icon_name = string_dup ("xdmchoose");
+	if (!XStringListToTextProperty (&caption, 1, &x_caption)) {
+		xc_ptr = NULL;
+	} else {
+		xc_ptr = &x_caption;
+	}
+	if (!XStringListToTextProperty (&icon_name, 1, &x_icon_name)) {
+		xi_ptr = NULL;
+	} else {
+		xi_ptr = &x_icon_name;
+	}
+	size_hints.flags = PSize | PMinSize | PMaxSize | PPosition;
+	size_hints.width = size_hints.min_width = size_hints.max_width = width;
+	size_hints.height = size_hints.min_height = size_hints.max_height = height;
+	size_hints.x = x;
+	size_hints.y = y;
+	wm_hints.flags = StateHint | InputHint;
+	wm_hints.initial_state = NormalState;
+	wm_hints.input = True;
+	app_name = string_dup ("xdmchoose");
+	app_class = string_dup ("xdmchoose");
+	class_hints.res_name = app_name;
+	class_hints.res_class = app_class;
+	XSetWMProperties (cxt->display, gadget->window, xc_ptr, xi_ptr, cxt->argv, cxt->argc, &size_hints, &wm_hints, &class_hints);
 	XMapWindow (cxt->display, gadget->window);
 	return gadget;
 }
-
-
+/*}}}*/
+/*{{{  void destroy_toplevel (Gadget *gadget)*/
 /*
  *	void destroy_toplevel (Gadget *gadget)
  *	Destroys a top-level gadget
@@ -171,8 +229,8 @@ void destroy_toplevel (Gadget *gadget)
 	destroy_gadget (gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  Gadget *create_panel (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height)*/
 /*
  *	Gadget *create_panel (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height)
  *	Creates a new panel gadget
@@ -210,10 +268,14 @@ Gadget *create_panel (MyXContext *cxt, char *name, Gadget *parent, int x, int y,
 	gadget->expose_event = panel_expose_event;
 	if (use_pixmap) {
 		if (gadget->bg_pixmap == None) {
+#ifdef USE_LIBXPM
 			i = XpmReadFileToPixmap (cxt->display, RootWindow (cxt->display, cxt->screen), gadget->config->background, &gadget->bg_pixmap, NULL, NULL);
 			if (i != XpmSuccess) {
 				gadget->bg_pixmap = None;
 			}
+#else
+			gadget->bg_pixmap = None;
+#endif
 		}
 		if (gadget->bg_pixmap != None) {
 			XSetWindowBackgroundPixmap (cxt->display, gadget->window, gadget->bg_pixmap);
@@ -222,7 +284,8 @@ Gadget *create_panel (MyXContext *cxt, char *name, Gadget *parent, int x, int y,
 	return gadget;
 }
 
-
+/*}}}*/
+/*{{{  void destroy_panel (Gadget *gadget)*/
 /*
  *	void destroy_panel (Gadget *gadget)
  *	Destroys a panel gadget
@@ -239,8 +302,8 @@ void destroy_panel (Gadget *gadget)
 	destroy_gadget (gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  Gadget *create_label (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, char *text)*/
 /*
  *	Gadget *create_label (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, char *text)
  *	Creates a new label gadget
@@ -294,8 +357,8 @@ Gadget *create_label (MyXContext *cxt, char *name, Gadget *parent, int x, int y,
 	}
 	return gadget;
 }
-
-
+/*}}}*/
+/*{{{  void destroy_label (Gadget *gadget)*/
 /*
  *	void destroy_label (Gadget *gadget)
  *	Destroys a label gadget
@@ -310,13 +373,13 @@ void destroy_label (Gadget *gadget)
 	destroy_gadget (gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int text_max, void (*accept_fcn)(Gadget *, char *), void (*escape_fcn)(Gadget *, char *), void (*ctrlc_fcn)(Gadget *, char *))*/
 /*
  *	Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int text_max, void (*accept_fcn)(Gadget *, char*))
  *	Creates a new edit gadget
  */
-Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int text_max, void (*accept_fcn)(Gadget *, char *), void (*escape_fcn)(Gadget *, char *), void (*ctrlc_fcn)(Gadget *, char *))
+Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int text_max, int allow_type, void (*accept_fcn)(Gadget *, char *), void (*escape_fcn)(Gadget *, char *), void (*ctrlc_fcn)(Gadget *, char *))
 {
 	Gadget *gadget;
 	XGCValues gc_values;
@@ -350,6 +413,7 @@ Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y,
 	pdata->max_len = text_max;
 	pdata->exposed = 0;
 	pdata->text_width = 0;
+	pdata->allow_type = allow_type;
 	pdata->edit_accept = accept_fcn;
 	pdata->edit_escape = escape_fcn;
 	pdata->edit_ctrlc = ctrlc_fcn;
@@ -361,8 +425,8 @@ Gadget *create_edit (MyXContext *cxt, char *name, Gadget *parent, int x, int y,
 	gadget->keypress_event = edit_keypress_event;
 	return gadget;
 }
-
-
+/*}}}*/
+/*{{{  void destroy_edit (Gadget *gadget)*/
 /*
  *	void destroy_edit (Gadget *gadget)
  *	destroys an edit gadget
@@ -377,8 +441,8 @@ void destroy_edit (Gadget *gadget)
 	destroy_gadget (gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  char *text_of_edit (Gadget *gadget)*/
 /*
  *	char *text_of_edit (Gadget *gadget)
  *	Returns the text held within and edit gadget
@@ -390,8 +454,34 @@ char *text_of_edit (Gadget *gadget)
 	pdata = (EditPrivate *)gadget->pdata;
 	return pdata->text;
 }
+/*}}}*/
+/*{{{  void settext_of_edit (Gadget *gadget, char *text)*/
+/*
+ *	Sets the text of an edit box
+ */
+void settext_of_edit (Gadget *gadget, char *text)
+{
+	EditPrivate *pdata;
+	int n;
 
-
+	pdata = (EditPrivate *)gadget->pdata;
+	n = strlen (text);
+	if (n > pdata->max_len) {
+		n = pdata->max_len;
+	}
+	strncpy (pdata->text, text, n);
+	pdata->text[n] = '\0';
+	pdata->cur_len = n;
+	pdata->cur_pos = n;
+	pdata->text_width = XTextWidth (gadget->fontinfo, pdata->text, pdata->cur_pos);
+	if (!pdata->exposed) {
+		return;
+	}
+	edit_expose_event (NULL, gadget);
+	return;
+}
+/*}}}*/
+/*{{{  Gadget *create_button (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, char *text, void (*button_fcn)(Gadget *))*/
 /*
  *	Gadget *create_button (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, char *text, void (*button_fcn)(Gadget *))
  *	Creates a new button gadget
@@ -401,10 +491,19 @@ Gadget *create_button (MyXContext *cxt, char *name, Gadget *parent, int x, int y
 	Gadget *gadget;
 	XGCValues gc_values;
 	ButtonPrivate *pdata;
+	int havetext = 0;
 
 	gadget = create_gadget (cxt, name, x, y, width, height);
-	if (read_config_entry (name, gadget->config, FLD_FOREGROUND | FLD_BACKGROUND | FLD_HIGHLIGHT | FLD_SHADOW | FLD_FONT | FLD_TEXT)) {
+	if (have_config_entry (name, FLD_TEXT) == 0) {
+		havetext = 1;
+	}
+	if (read_config_entry (name, gadget->config, FLD_FOREGROUND | FLD_BACKGROUND | FLD_HIGHLIGHT | FLD_SHADOW | (havetext ? (FLD_TEXT | FLD_FONT) : 0))) {
 		return NULL;
+	} else if (!havetext && !text) {
+		return NULL;
+	}
+	if (!havetext && text) {
+		gadget->config->text = string_dup (text);
 	}
 	pdata = (ButtonPrivate *)xmalloc (sizeof (ButtonPrivate));
 	allocate_color (cxt, gadget->config->foreground, &gadget->fg_pixel);
@@ -412,20 +511,36 @@ Gadget *create_button (MyXContext *cxt, char *name, Gadget *parent, int x, int y
 	allocate_color (cxt, gadget->config->highlight, &gadget->hi_pixel);
 	allocate_color (cxt, gadget->config->shadow, &gadget->sh_pixel);
 	gadget->window = XCreateSimpleWindow (cxt->display, parent->window, x, y, width, height, 0, BlackPixel (cxt->display, cxt->screen), gadget->bg_pixel);
-	gadget->fontinfo = XLoadQueryFont (cxt->display, gadget->config->font);
-	if (!gadget->fontinfo) {
-		gadget->fontinfo = XLoadQueryFont (cxt->display, "fixed");
+	if (!strcmp (text, "_")) {
+		pdata->doarrow = 1;
+		pdata->arrow_points[0].x = (width >> 1) - 4;
+		pdata->arrow_points[0].y = (height >> 1) - 4;
+		pdata->arrow_points[1].x = (width >> 1) + 4;
+		pdata->arrow_points[1].y = (height >> 1) - 4;
+		pdata->arrow_points[2].x = width >> 1;
+		pdata->arrow_points[2].y = (height >> 1) + 4;
+	} else {
+		pdata->doarrow = 0;
+		gadget->fontinfo = XLoadQueryFont (cxt->display, gadget->config->font);
 		if (!gadget->fontinfo) {
-			return NULL;
+			gadget->fontinfo = XLoadQueryFont (cxt->display, "fixed");
+			if (!gadget->fontinfo) {
+				return NULL;
+			}
 		}
+		gc_values.font = gadget->fontinfo->fid;
 	}
 	gc_values.background = gadget->bg_pixel;
-	gc_values.font = gadget->fontinfo->fid;
-	gadget->gc = XCreateGC (cxt->display, gadget->window, GCBackground | GCFont, &gc_values);
-	pdata->text_x = (gadget->width >> 1) - (XTextWidth (gadget->fontinfo, gadget->config->text, strlen (gadget->config->text)) >> 1);
-	pdata->text_y = (gadget->height >> 1) + (gadget->fontinfo->ascent >> 1);
-	pdata->text = gadget->config->text;
-	pdata->len = strlen (pdata->text);
+	gadget->gc = XCreateGC (cxt->display, gadget->window, GCBackground | (pdata->doarrow ? 0 : GCFont), &gc_values);
+	if (!pdata->doarrow) {
+		pdata->text_x = (gadget->width >> 1) - (XTextWidth (gadget->fontinfo, gadget->config->text, strlen (gadget->config->text)) >> 1);
+		pdata->text_y = (gadget->height >> 1) + (gadget->fontinfo->ascent >> 1);
+		pdata->text = gadget->config->text;
+		pdata->len = strlen (pdata->text);
+	} else {
+		pdata->text = gadget->config->text;
+		pdata->len = 0;
+	}
 	pdata->mouse_in = 0;
 	pdata->mouse_down = 0;
 	pdata->exposed = 0;
@@ -440,8 +555,8 @@ Gadget *create_button (MyXContext *cxt, char *name, Gadget *parent, int x, int y
 	gadget->mouseleave_event = button_mouseleave_event;
 	return gadget;
 }
-
-
+/*}}}*/
+/*{{{  void destroy_button (Gadget *gadget)*/
 /*
  *	void destroy_button (Gadget *gadget)
  *	Destroys a button gadget
@@ -456,8 +571,141 @@ void destroy_button (Gadget *gadget)
 	destroy_gadget (gadget);
 	return;
 }
+/*}}}*/
+/*{{{  Gadget *create_listbox (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int visible, char **items, Gadget *below, void (*select_fcn)(Gadget *, int))*/
+/*
+ *	Creates a new list-box gadget
+ */
+Gadget *create_listbox (MyXContext *cxt, char *name, Gadget *parent, int x, int y, int width, int height, int visible, char **items, Gadget *below, void (*select_fcn)(Gadget *, int))
+{
+	Gadget *gadget;
+	XGCValues gc_values;
+	ListPrivate *pdata;
+	int i;
+	XSetWindowAttributes win_attr;
 
+	gadget = create_gadget (cxt, name, x, y, width, height);
+	if (read_config_entry (name, gadget->config, FLD_FOREGROUND | FLD_BACKGROUND | FLD_HIGHLIGHT | FLD_SHADOW | FLD_FONT)) {
+		return NULL;
+	}
+	pdata = (ListPrivate *)xmalloc (sizeof (ListPrivate));
+	allocate_color (cxt, gadget->config->foreground, &gadget->fg_pixel);
+	allocate_color (cxt, gadget->config->background, &gadget->bg_pixel);
+	allocate_color (cxt, gadget->config->highlight, &gadget->hi_pixel);
+	allocate_color (cxt, gadget->config->shadow, &gadget->sh_pixel);
+	gadget->fontinfo = XLoadQueryFont (cxt->display, gadget->config->font);
+	if (!gadget->fontinfo) {
+		gadget->fontinfo = XLoadQueryFont (cxt->display, "fixed");
+		if (!gadget->fontinfo) {
+			return NULL;
+		}
+	}
+	gc_values.background = gadget->bg_pixel;
+	gc_values.font = gadget->fontinfo->fid;
+	pdata->options = items;
+	for (i=0; items[i]; i++);
+	pdata->n_opts = i;
+	pdata->exposed = 0;
+	pdata->text_height = gadget->fontinfo->ascent + gadget->fontinfo->descent + 4;
+	pdata->text_yoffs = gadget->fontinfo->ascent;
+	pdata->below = below;
+	pdata->pop_x = x;
+	pdata->pop_y = y;
+	pdata->grabbed = 0;
+	pdata->list_select = select_fcn;
+	pdata->g_cur = XCreateFontCursor (cxt->display, XC_arrow);
+	height = (i * pdata->text_height) + 4;
+	gadget->height = height;
 
+	win_attr.background_pixel = gadget->bg_pixel;
+	win_attr.save_under = True;
+	win_attr.override_redirect = True;
+	gadget->window = XCreateWindow (cxt->display, RootWindow (cxt->display, cxt->screen), pdata->pop_x, pdata->pop_y, width, height, 0,
+				DefaultDepth (cxt->display, cxt->screen), InputOutput, DefaultVisual (cxt->display, cxt->screen),
+				CWBackPixel | CWSaveUnder | CWOverrideRedirect, &win_attr);
+
+	/* gadget->window = XCreateSimpleWindow (cxt->display, parent->window, x, y, width, height, 0, BlackPixel (cxt->display, cxt->screen), gadget->bg_pixel); */
+	gadget->gc = XCreateGC (cxt->display, gadget->window, GCBackground | GCFont, &gc_values);
+	gadget->visible = visible;
+	/* creating a list box with NULL items is a bad idea.. */
+
+	XSelectInput (cxt->display, gadget->window, ExposureMask | ButtonPressMask | ButtonReleaseMask | EnterWindowMask | LeaveWindowMask);
+	if (visible) {
+		XMapWindow (cxt->display, gadget->window);
+	}
+	gadget->pdata = (void *)pdata;
+	gadget->expose_event = list_expose_event;
+	gadget->mousedown_event = list_mousedown_event;
+	gadget->mouseup_event = list_mouseup_event;
+	return gadget;
+}
+/*}}}*/
+/*{{{  void destroy_listbox (Gadget *gadget)*/
+/*
+ *	Destroys a list-box
+ */
+void destroy_listbox (Gadget *gadget)
+{
+	if (gadget->visible) {
+		XUnmapWindow (gadget->xcxt->display, gadget->window);
+	}
+	XDestroyWindow (gadget->xcxt->display, gadget->window);
+	XFreeGC (gadget->xcxt->display, gadget->gc);
+	XFreeFont (gadget->xcxt->display, gadget->fontinfo);
+	free (gadget->pdata);
+	destroy_gadget (gadget);
+	return;
+}
+/*}}}*/
+/*{{{  void show_listbox (Gadget *gadget)*/
+/*
+ *	makes the listbox visible
+ */
+void show_listbox (Gadget *gadget)
+{
+	ListPrivate *pdata;
+
+	pdata = (ListPrivate *)(gadget->pdata);
+	if (gadget->visible) {
+		return;
+	}
+	list_recalc_xypos (gadget);
+	XMoveWindow (gadget->xcxt->display, gadget->window, pdata->pop_x, pdata->pop_y);
+	gadget->x = pdata->pop_x;
+	gadget->y = pdata->pop_y;
+	XMapWindow (gadget->xcxt->display, gadget->window);
+	XFlush (gadget->xcxt->display);
+	pdata->grabbed = 0;
+	if (XGrabPointer (gadget->xcxt->display, gadget->window, False, ButtonPressMask | ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, pdata->g_cur, CurrentTime)) {
+		pdata->grabbed = 1;
+	}
+	gadget->visible = 1;
+	return;
+}
+/*}}}*/
+/*{{{  void hide_listbox (Gadget *gadget)*/
+/*
+ *	hides the listbox
+ */
+void hide_listbox (Gadget *gadget)
+{
+	ListPrivate *pdata;
+
+	pdata = (ListPrivate *)(gadget->pdata);
+	if (!gadget->visible) {
+		return;
+	}
+	if (pdata->grabbed) {
+		XUngrabPointer (gadget->xcxt->display, CurrentTime);
+		pdata->grabbed = 0;
+	}
+	XUnmapWindow (gadget->xcxt->display, gadget->window);
+	gadget->visible = 0;
+	return;
+}
+/*}}}*/
+
+/*{{{  static void allocate_color (MyXContext *cxt, char *colordef, unsigned long *xpixel)*/
 /*
  *	void allocate_color (MyXContext *cxt, char *colordef, unsigned long *xpixel)
  *	Allocates a color
@@ -473,8 +721,8 @@ static void allocate_color (MyXContext *cxt, char *colordef, unsigned long *xpix
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void panel_expose_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void panel_expose_event (XEvent *ev, Gadget *gadget)
  *	Called when a panel gets exposed
@@ -496,8 +744,8 @@ static void panel_expose_event (XEvent *ev, Gadget *gadget)
 	XDrawLines (gadget->xcxt->display, gadget->window, gadget->gc, points, 3, CoordModeOrigin);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void label_expose_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void label_expose_event (XEvent *ev, Gadget *gadget)
  *	Called when a label gets exposed
@@ -512,8 +760,8 @@ static void label_expose_event (XEvent *ev, Gadget *gadget)
 	XDrawString (gadget->xcxt->display, gadget->window, gadget->gc, pdata->text_x, pdata->text_y, pdata->text, pdata->len);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void label_show_action (Gadget *gadget)*/
 /*
  *	void label_show_action (Gadget *gadget)
  *	Shows a label
@@ -526,8 +774,8 @@ static void label_show_action (Gadget *gadget)
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void label_hide_action (Gadget *gadget)*/
 /*
  *	void label_hide_action (Gadget *gadget)
  *	Hides a label
@@ -540,8 +788,8 @@ static void label_hide_action (Gadget *gadget)
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void label_settext_action (Gadget *gadget, char *text)*/
 /*
  *	void label_settext_action (Gadget *gadget, char *text)
  *	Sets the text of the label
@@ -565,8 +813,8 @@ static void label_settext_action (Gadget *gadget, char *text)
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void edit_expose_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void edit_expose_event (XEvent *ev, Gadget *gadget)
  *	Called when an edit gadget gets exposed
@@ -597,8 +845,8 @@ static void edit_expose_event (XEvent *ev, Gadget *gadget)
 	XDrawLine (gadget->xcxt->display, gadget->window, gadget->gc, pdata->text_width + 4, 4, pdata->text_width + 4, gadget->height - 4);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void edit_keypress_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void edit_keypress_event (XEvent *ev, Gadget *gadget)
  *	Called when a keypress happens on an edit gadget
@@ -634,7 +882,9 @@ static void edit_keypress_event (XEvent *ev, Gadget *gadget)
 			}
 		} else if (((keysym>=XK_KP_Enter) && (keysym<=XK_KP_9)) || ((keysym>=XK_space) && (keysym<=XK_asciitilde)) || ((keysym >= XK_F1) && (keysym <= XK_F35))) {
 			blen = strlen (buffer);
-			if (((pdata->cur_len + blen) >= pdata->max_len) || !blen) {
+			if (!pdata->allow_type) {
+				XBell (gadget->xcxt->display, 0);
+			} else if (((pdata->cur_len + blen) >= pdata->max_len) || !blen) {
 				XBell (gadget->xcxt->display, 0);
 			} else {
 				if (pdata->cur_pos == pdata->cur_len) {
@@ -657,7 +907,9 @@ static void edit_keypress_event (XEvent *ev, Gadget *gadget)
 		} else if ((keysym >= XK_Shift_L) && (keysym <= XK_Hyper_R)) {
 			/* Modifier */
 		} else if ((keysym==XK_BackSpace) || (keysym==XK_Delete)) {
-			if (pdata->cur_pos > 0) {
+			if (!pdata->allow_type) {
+				XBell (gadget->xcxt->display, 0);
+			} else if (pdata->cur_pos > 0) {
 				if (pdata->cur_pos == pdata->cur_len) {
 					pdata->text[pdata->cur_pos - 1] = '\0';
 				} else {
@@ -704,8 +956,8 @@ static void edit_keypress_event (XEvent *ev, Gadget *gadget)
 	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void button_expose_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void button_expose_event (XEvent *ev, Gadget *gadget)
  *	Called when button is exposed
@@ -741,11 +993,19 @@ static void button_expose_event (XEvent *ev, Gadget *gadget)
 	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->bg_pixel);
 	XFillRectangle (gadget->xcxt->display, gadget->window, gadget->gc, 1, 1, gadget->width - 2, gadget->height - 2);
 	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->fg_pixel);
-	XDrawImageString (gadget->xcxt->display, gadget->window, gadget->gc, pdata->text_x + disp, pdata->text_y + disp, pdata->text, pdata->len);
+	if (pdata->doarrow) {
+		XPoint lpnts[3] = {{pdata->arrow_points[0].x, pdata->arrow_points[0].y + disp},
+				   {pdata->arrow_points[1].x, pdata->arrow_points[1].y + disp},
+				   {pdata->arrow_points[2].x, pdata->arrow_points[2].y + disp}};
+
+		XFillPolygon (gadget->xcxt->display, gadget->window, gadget->gc, lpnts, 3, Convex, CoordModeOrigin);
+	} else {
+		XDrawImageString (gadget->xcxt->display, gadget->window, gadget->gc, pdata->text_x + disp, pdata->text_y + disp, pdata->text, pdata->len);
+	}
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void button_mousedown_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void button_mousedown_event (XEvent *ev, Gadget *gadget)
  *	Called when mouse button pressed on button
@@ -762,8 +1022,8 @@ static void button_mousedown_event (XEvent *ev, Gadget *gadget)
 	button_expose_event (ev, gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void button_mouseup_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void button_mouseup_event (XEvent *ev, Gadget *gadget)
  *	Called when mouse button released
@@ -783,8 +1043,8 @@ static void button_mouseup_event (XEvent *ev, Gadget *gadget)
 	button_expose_event (ev, gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void button_mouseenter_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void button_mouseenter_event (XEvent *ev, Gadget *gadget)
  *	Called when mouse enters button
@@ -801,8 +1061,8 @@ static void button_mouseenter_event (XEvent *ev, Gadget *gadget)
 	button_expose_event (ev, gadget);
 	return;
 }
-
-
+/*}}}*/
+/*{{{  static void button_mouseleave_event (XEvent *ev, Gadget *gadget)*/
 /*
  *	void button_mouseleave_event (XEvent *ev, Gadget *gadget)
  *	Called when mouse button leaves
@@ -819,8 +1079,115 @@ static void button_mouseleave_event (XEvent *ev, Gadget *gadget)
 	button_expose_event (ev, gadget);
 	return;
 }
+/*}}}*/
+/*{{{  static void list_expose_event (XEvent *ev, Gadget *gadget)*/
+static void list_expose_event (XEvent *ev, Gadget *gadget)
+{
+	ListPrivate *pdata;
+	XPoint points[4];
+	int i;
 
+	pdata = (ListPrivate *)gadget->pdata;
+	pdata->exposed = 1;
 
+	if ((gadget->x != pdata->pop_x) || (gadget->y != pdata->pop_y)) {
+		XMoveWindow (gadget->xcxt->display, gadget->window, pdata->pop_x, pdata->pop_y);
+		gadget->x = pdata->pop_x;
+		gadget->y = pdata->pop_y;
+	}
 
+	points[0].x = 0; points[0].y = (gadget->height - 1);
+	points[1].x = 0; points[1].y = 0;
+	points[2].x = (gadget->width - 1); points[2].y = 0;
+	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->sh_pixel);
+	XDrawLines (gadget->xcxt->display, gadget->window, gadget->gc, points, 3, CoordModeOrigin);
+	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->hi_pixel);
+	points[0].x = 1; points[0].y = (gadget->height - 1);
+	points[1].x = (gadget->width - 1); points[1].y = (gadget->height - 1);
+	points[2].x = (gadget->width - 1); points[2].y = 1;
+	points[3].x = 0; points[3].y = 1;
+	XDrawLines (gadget->xcxt->display, gadget->window, gadget->gc, points, 4, CoordModeOrigin);
+	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->bg_pixel);
+	XFillRectangle (gadget->xcxt->display, gadget->window, gadget->gc, 1, 2, gadget->width - 2, gadget->height - 3);
+	XSetForeground (gadget->xcxt->display, gadget->gc, gadget->fg_pixel);
+	/* draw options */
+	for (i=0; i<pdata->n_opts; i++) {
+		XDrawImageString (gadget->xcxt->display, gadget->window, gadget->gc, 4, (pdata->text_height * i) + (2 + pdata->text_yoffs), pdata->options[i], strlen (pdata->options[i]));
+	}
+	return;
+}
+/*}}}*/
+/*{{{  static void list_mousedown_event (XEvent *ev, Gadget *gadget)*/
+static void list_mousedown_event (XEvent *ev, Gadget *gadget)
+{
+	ListPrivate *pdata;
+	int y, x, i;
 
+	pdata = (ListPrivate *)gadget->pdata;
+	if (!pdata->exposed) {
+		return;
+	}
+	y = ev->xbutton.y;
+	x = ev->xbutton.x;
+	
+	/* we get events, see if its outside */
+	if ((y < 0) || (x < 0) || (y >= gadget->height) || (x >= gadget->width)) {
+		hide_listbox (gadget);
+		return;
+	}
+	i = y / pdata->text_height;
+	if (pdata->list_select && (i < pdata->n_opts)) {
+		pdata->list_select (gadget, i);
+	}
+	hide_listbox (gadget);
+	return;
+}
+/*}}}*/
+/*{{{  static void list_mouseup_event (XEvent *ev, Gadget *gadget)*/
+static void list_mouseup_event (XEvent *ev, Gadget *gadget)
+{
+	/* nothing.. */
+	return;
+}
+/*}}}*/
+/*{{{  static void list_recalc_xypos (Gadget *gadget)*/
+/*
+ *	re-calculates the location of the popup window, based on
+ *	its "below" gadget (for xdmchooser, the edit box)
+ */
+static void list_recalc_xypos (Gadget *gadget)
+{
+	Window root, r_root;
+	Window w, parent;
+	Window *clist;
+	int n_children;
+	ListPrivate *pdata;
+
+	pdata = (ListPrivate *)gadget->pdata;
+	w = pdata->below->window;
+	root = RootWindow (gadget->xcxt->display, gadget->xcxt->screen);
+
+	pdata->pop_x = 0;
+	pdata->pop_y = pdata->below->height - 1;
+	while (w != root) {
+		int rx, ry;
+		unsigned int rw, rh, rbw, rd;
+
+		if (!XGetGeometry (gadget->xcxt->display, (Drawable)w, &r_root, &rx, &ry, &rw, &rh, &rbw, &rd)) {
+			return;
+		}
+		pdata->pop_x += rx + rbw;
+		pdata->pop_y += ry + rbw;
+		root = r_root;			/* root should not change, but..! */
+		if (!XQueryTree (gadget->xcxt->display, w, &r_root, &parent, &clist, &n_children)) {
+			return;
+		}
+		if (clist) {
+			XFree (clist);		/* not interested */
+		}
+		w = parent;
+	}
+	return;
+}
+/*}}}*/
 
