@@ -20,7 +20,11 @@
 #ifndef __VERSION_H
 #define __VERISON_H
 
-#define XDMCHOOSE_VERSION	"1.1.2"
+#ifdef VERSION
+#define XDMCHOOSE_VERSION VERSION
+#else
+#define XDMCHOOSE_VERSION "(1.2.1)"
+#endif
 
 #endif	/* !__VERSION_H */
 
