@@ -43,7 +43,7 @@ GlobalConfigData _globconfig;
 
 /* Statics */
 
-#define MAX_NUM_GADGETS (10)
+#define MAX_NUM_GADGETS (15)
 static int num_gadgets;
 static Gadget *gadgets[MAX_NUM_GADGETS];
 static Gadget *g_toplevel;
@@ -57,6 +57,9 @@ static Gadget *g_loginbutton;
 static Gadget *g_querybutton;
 static Gadget *g_copylabel;
 static Gadget *g_extratext;
+static Gadget *g_combobtn;
+static Gadget *g_hostlist;
+
 static int mainloop_running, mainloop_termcode;
 static void (*pp_select_function)(char *);
 static void (*pp_check_function)(char *);
@@ -73,6 +76,8 @@ static void g_hostedit_ctrlc (Gadget *gadget, char *string);
 static void g_loginbutton_click (Gadget *gadget);
 static void g_querybutton_click (Gadget *gadget);
 static int kill_window (char *window_name, Window start_at, int depth);
+static void g_combobtn_click (Gadget *gadget);
+static void g_hostlist_select (Gadget *gadget, int idx);
 
 
 /* Own prototypes */
@@ -121,8 +126,8 @@ int init_xparts (char **in_argv, int in_argc)
 	gadgets[num_gadgets++] = g_panel = create_panel (&_xcontext, "panel", g_toplevel, 0, 0, xdim, ydim);
 	gadgets[num_gadgets++] = g_title = create_label (&_xcontext, "title", g_panel, 1, 1, xdim-2, 48, "X Window System");
 	gadgets[num_gadgets++] = g_hostlabel = create_label (&_xcontext, "hostlabel", g_panel, 96, 66, 48, 24, "Host :");
-	gadgets[num_gadgets++] = g_hostedit = create_edit (&_xcontext, "hostedit", g_panel, 160, 64, 128, 28, 128, g_hostedit_accept, g_hostedit_escape, g_hostedit_ctrlc);
-	gadgets[num_gadgets++] = g_copylabel = create_label (&_xcontext, "copylabel", g_panel, 1, 163, xdim-2, 16, "Copyright (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>");
+	gadgets[num_gadgets++] = g_hostedit = create_edit (&_xcontext, "hostedit", g_panel, 160, 64, 128, 28, 128, _globconfig.allowtype, g_hostedit_accept, g_hostedit_escape, g_hostedit_ctrlc);
+	gadgets[num_gadgets++] = g_copylabel = create_label (&_xcontext, "copylabel", g_panel, 1, 163, xdim-2, 16, "XDM-Choose " VERSION ".  Copyright (C) 2000-2003 Fred Barnes <fred@frmb.org>");
 	gadgets[num_gadgets++] = g_errorlabel = create_label (&_xcontext, "errorlabel", g_panel, 1, 96, xdim-2, 24, "");
 	g_errorlabel->hide_action (g_errorlabel);
 	gadgets[num_gadgets++] = g_statuslabel = create_label (&_xcontext, "statuslabel", g_panel, 1, 96, xdim-2, 24, "");
@@ -138,6 +143,11 @@ int init_xparts (char **in_argv, int in_argc)
 		gadgets[num_gadgets++] = g_extratext = create_label (&_xcontext, "extratext", g_panel, 1,
 			g_copylabel->y + (g_copylabel->fontinfo->ascent + g_copylabel->fontinfo->descent) + 4, xdim-2, 16, "");
 	}
+	if (_globconfig.hostlist) {
+		/* offering drop-down selection */
+		gadgets[num_gadgets++] = g_combobtn = create_button (&_xcontext, "combobutton", g_panel, 290, 66, 16, 24, "_", g_combobtn_click);
+		gadgets[num_gadgets++] = g_hostlist = create_listbox (&_xcontext, "hostlist", g_panel, 160, 91, 128, 128, 0, _globconfig.hostlist, g_hostedit, g_hostlist_select);
+	}
 	return 0;
 }
 
@@ -253,10 +263,12 @@ int xparts_mainloop (void)
 	}
 	mainloop_running = 1;
 	mainloop_termcode = 0;
-	result = XGrabKeyboard (_xcontext.display, g_hostedit->window, False, GrabModeAsync, GrabModeAsync, CurrentTime);
-	if (result) {
-		ERR_MSG ("Unable to XGrabKeyboard ()");
-		return -1;
+	if (_globconfig.grabkybd) {
+		result = XGrabKeyboard (_xcontext.display, g_hostedit->window, False, GrabModeAsync, GrabModeAsync, CurrentTime);
+		if (result) {
+			ERR_MSG ("Unable to XGrabKeyboard ()");
+			return -1;
+		}
 	}
 	while (mainloop_running) {
 		if ((xp_fd > -1) && !XPending (_xcontext.display)) {
@@ -319,6 +331,11 @@ int xparts_mainloop (void)
 		case KeyRelease:
 			if (gadgets[i]->keypress_event) {
 				gadgets[i]->keypress_event (&ev, gadgets[i]);
+			} else if (gadgets[i] == g_toplevel) {
+				/* probably means there's a window-manager or something in the way.. */
+				if (g_hostedit && g_hostedit->keypress_event) {
+					g_hostedit->keypress_event (&ev, g_hostedit);
+				}
 			}
 			break;
 		case EnterNotify:
@@ -335,7 +352,9 @@ int xparts_mainloop (void)
 			break;
 		}
 	}
-	XUngrabKeyboard (_xcontext.display, CurrentTime);
+	if (_globconfig.grabkybd) {
+		XUngrabKeyboard (_xcontext.display, CurrentTime);
+	}
 	return mainloop_termcode;
 }
 
@@ -367,6 +386,7 @@ void set_error_message (char *message)
 	g_errorlabel->settext_action (g_errorlabel, message);
 	g_statuslabel->hide_action (g_statuslabel);
 	g_errorlabel->show_action (g_errorlabel);
+	XFlush (_xcontext.display);
 	return;
 }
 
@@ -380,6 +400,7 @@ void set_status_message (char *message)
 	g_statuslabel->settext_action (g_statuslabel, message);
 	g_errorlabel->hide_action (g_errorlabel);
 	g_statuslabel->show_action (g_statuslabel);
+	XFlush (_xcontext.display);
 	return;
 }
 
@@ -481,4 +502,29 @@ static void g_querybutton_click (Gadget *gadget)
 }
 
 
+/*
+ *	void g_combobtn_click (Gadget *gadget)
+ *	called when host-list combo button thing is clicked
+ *	this event happens as soon as the button is pressed!
+ */
+static void g_combobtn_click (Gadget *gadget)
+{
+	if (!g_hostlist->visible) {
+		show_listbox (g_hostlist);
+	} else {
+		hide_listbox (g_hostlist);
+	}
+	return;
+}
+
+
+/*
+ *	void g_hostlist_select (Gadget *gadget, int idx)
+ *	called when a host is selected from the (combo) list
+ */
+static void g_hostlist_select (Gadget *gadget, int idx)
+{
+	settext_of_edit (g_hostedit, _globconfig.hostlist[idx]);
+	return;
+}
 
